/*
 * 
 *TFM-TetrisLite
 *Master en Ingenieria Informatica
 *Julio Fernandez Lasheras
 *02/01/2017
 *
 */


// Include application, user and local libraries
#include "SPI.h"
#include "Energia.h"

#include "joystick.h"
#include "music.h"

#include "Screen_HX8353E.h"
Screen_HX8353E myScreen;

////////////////////////////Defines//////////////////////////////////////
#define LCD_WIDTH             120
#define LCD_HEIGHT            120

#define MIN(X, Y)             (((X) < (Y)) ? (X) : (Y))

#define BOARD_WIDTH           9
#define BOARD_HEIGHT          16

#define BLOCK_SIZE            MIN( (LCD_WIDTH-1) / BOARD_WIDTH, (LCD_HEIGHT-1) / BOARD_HEIGHT )

#define BOARD_LEFT            (LCD_WIDTH - BOARD_WIDTH * BLOCK_SIZE)/4*3
#define BOARD_RIGHT           (BOARD_LEFT + BLOCK_SIZE * BOARD_WIDTH)
#define BOARD_TOP             (LCD_HEIGHT - BOARD_HEIGHT * BLOCK_SIZE) / 2
#define BOARD_BOTTOM          (BOARD_TOP + BOARD_HEIGHT * BLOCK_SIZE)
#define newBlockx             5
#define newBlocky             70

#define BUTTON_A              33
#define BUTTON_B              32

#define BLUE_LED              37

#define DROP_WAIT_INIT_0      1100
#define DROP_WAIT_INIT_1      800
#define DROP_WAIT_INIT_2      600
#define DROP_WAIT_INIT_3      500

#define INPUT_WAIT_ROT        700
#define INPUT_WAIT_MOVE       300

#define INPUT_WAIT_NEW_SHAPE  400

uint8_t lShape[4][4][2] = {
  {{0, 0}, {0, 1}, {0, 2}, {1, 2}},
  {{0, 1}, {1, 1}, {2, 0}, {2, 1}},
  {{0, 0}, {1, 0}, {1, 1}, {1, 2}},
  {{0, 0}, {0, 1}, {1, 0}, {2, 0}},
};

uint8_t jShape[4][4][2] = {
  {{1, 0}, {1, 1}, {0, 2}, {1, 2}},
  {{0, 0}, {1, 0}, {2, 0}, {2, 1}},
  {{0, 0}, {1, 0}, {0, 1}, {0, 2}},
  {{0, 0}, {0, 1}, {1, 1}, {2, 1}},
};

uint8_t o_shape[1][4][2] = {
  { {0, 0}, {0, 1}, {1, 0}, {1, 1}}
};

uint8_t s_shape[2][4][2] = {
  {{0, 1}, {1, 0}, {1, 1}, {2, 0}},
  {{0, 0}, {0, 1}, {1, 1}, {1, 2}}
};

uint8_t zShape[2][4][2] = {
  {{0, 0}, {1, 0}, {1, 1}, {2, 1}},
  {{1, 0}, {0, 1}, {1, 1}, {0, 2}}
};

uint8_t tShape[4][4][2] = {
  {{0, 0}, {1, 0}, {2, 0}, {1, 1}},
  {{0, 0}, {0, 1}, {1, 1}, {0, 2}},
  {{1, 0}, {0, 1}, {1, 1}, {2, 1}},
  {{1, 0}, {0, 1}, {1, 1}, {1, 2}}
};

uint8_t iShape[2][4][2] = {
  {{0, 0}, {1, 0}, {2, 0}, {3, 0}},
  {{0, 0}, {0, 1}, {0, 2}, {0, 3}}
};

// used to clear the position from the screen
struct Backup {
  uint8_t x, y, rot;
};

uint8_t shapes[7] = {4, 4, 1, 2, 2, 4, 2};
uint8_t *allShapes[7] = {lShape[0][0], jShape[0][0], o_shape[0][0], s_shape[0][0], zShape[0][0], tShape[0][0], iShape[0][0]};
unsigned int colors[7] = {orangeColour, blueColour, yellowColour, greenColour, redColour, magentaColour, cyanColour};

uint8_t current;
uint8_t nextSelected[7];
uint8_t nextShapeNumber;
unsigned long lastInput, lastDrop;
uint8_t board[BOARD_HEIGHT][BOARD_WIDTH];
uint8_t x, y, rot;
Backup old;
boolean newShape;
unsigned int dropWait;
volatile boolean state = false;
int lines, level, highScore = 0;
boolean isPaused = false;

unsigned long lastAPressed = 0;
unsigned long lastBPressed = 0;


///////////////////////////Function Prototypes/////////////////////////////
void chooseNewShape();
void randomShape();
void drawSelectedShape();
boolean checkEdge(int xPosition, int yPosition, int rotation);
void moveShapeDown();
void drawPreGameScreen();
void userInput(unsigned long now);
void gameOver();
void splashScreen(void);
void howToPlay(void);
void screenDraw(void);
void tetrisGame(void);
void lineCheck();
void pointsCount();
void pause();

void buttonCheck()
{
  unsigned long nowA = millis();
  //avoid debounce
  if (nowA - lastAPressed > 300 )
  {
    if (!state)
      state = true;
    else
      state = false;
    lastAPressed = nowA;
  }
}

void button2Check()
{
  unsigned long nowB = millis();
  //avoid debounce
  if (nowB - lastBPressed > 300 )
  {
    if (!isPaused)
      isPaused = true;
    else
      isPaused = false;
    lastBPressed = nowB;
  }
}

// Add setup code
void setup()
{
  pinMode(BUTTON_A, INPUT_PULLUP);
  attachInterrupt(BUTTON_A, buttonCheck, FALLING); // Interrupt is fired whenever button A is pressed
  pinMode(BUTTON_B, INPUT_PULLUP);
  attachInterrupt(BUTTON_B, button2Check, FALLING); // Interrupt is fired whenever button B is pressed
  
  myScreen.begin();
  myScreen.setFontSize(myScreen.fontMax());
  
  state = false;
  
  splashScreen();
  do{
    playTetrisTheme(); //play tetris theme while A button is not pressed
  }while (!state);

  restartTetrisTheme();
      
  howToPlay();
  // clear board

  for ( int i = 0; i < BOARD_WIDTH; i++ )
  {
    for ( int j = 0; j < BOARD_HEIGHT; j++)
    {
      board[j][i] = 0;
    }
  }
  current = 0;
  nextShapeNumber = 0;
  lastInput = 0;
  lastDrop = 0;
  lines = 0;
  level = 0;
  
  screenDraw();

}

// Add loop code
void loop()
{

  tetrisGame();
  pause();
  
}

void splashScreen(void)
{
  myScreen.clear(blackColour);

  myScreen.gText(10, 20, "TETRIS LITE", redColour);

  myScreen.gText(10, 50, "Press A button", redColour);
  myScreen.gText(10, 60, "to begin", redColour);

  myScreen.gText(10, 80, "High Score: ", redColour);
  myScreen.gText(10, 90, String(highScore), redColour);
}

void howToPlay(void)
{
  myScreen.clear(blackColour);

  myScreen.gText(10, 10, "HOW TO PLAY", whiteColour);

  myScreen.gText(10, 30, "Move with Joystick", whiteColour);
  myScreen.gText(10, 50, "Press A button", whiteColour);
  myScreen.gText(10, 60, "to Rotate", whiteColour);
  myScreen.gText(10, 80, "Press B button", whiteColour);
  myScreen.gText(10, 90, "to Pause", whiteColour);
  delay(2000);
}

void screenDraw(void)
{

  randomSeed(analogRead(0));

  myScreen.clear(blueColour);

  randomShape();

  lastInput = 0;
  lastDrop = 0;
  dropWait = DROP_WAIT_INIT_0;


  myScreen.setPenSolid(true);
  myScreen.dRectangle(BOARD_LEFT, BLOCK_SIZE /2 + 1, BOARD_RIGHT - BOARD_LEFT - 1, BOARD_BOTTOM - 5, blackColour);
  // Left Wall
  myScreen.line (
    BOARD_LEFT - 1,
    BOARD_TOP,
    BOARD_LEFT - 1,
    BOARD_BOTTOM,
    violetColour);

  // Right Wall

  myScreen.line (
    BOARD_RIGHT,
    BOARD_TOP,
    BOARD_RIGHT,
    BOARD_BOTTOM,
    violetColour);

  // Bottom Wall

  myScreen.line (
    BOARD_LEFT - 1,
    BOARD_BOTTOM,
    BOARD_RIGHT + 1,
    BOARD_BOTTOM,
    violetColour);
  //Horizontal Lines
  for ( int i = BOARD_LEFT + BLOCK_SIZE - 1; i < BOARD_RIGHT; i += BLOCK_SIZE)
  {
    myScreen.line (
      i,
      BOARD_TOP,
      i,
      BOARD_BOTTOM - 1,
      darkGrayColour);
  }
  //Vertical Lines
  for ( int i = BOARD_TOP + BLOCK_SIZE - 1; i < BOARD_BOTTOM; i += BLOCK_SIZE)
  {
    myScreen.line (
      BOARD_LEFT,
      i,
      BOARD_RIGHT - 1,
      i,
      darkGrayColour);
  }
  
  // draw next box
  myScreen.dRectangle(newBlockx - 1, newBlocky - 11, BLOCK_SIZE * 5 - 1, BLOCK_SIZE * 4 + 11, whiteColour);
  myScreen.gText(newBlockx, newBlocky - 10, "NEXT", blackColour, whiteColour);


  // draw score and level box
  myScreen.dRectangle(newBlockx - 1, newBlocky - 1 - 60, BLOCK_SIZE * 5 - 1, BLOCK_SIZE * 6 + 1, whiteColour);
  myScreen.gText(newBlockx, 10, "SCORE", blackColour, whiteColour);
  myScreen.gText(newBlockx + 2, 30, "LEVEL",  blackColour, whiteColour);
   
  pointsCount();
  newShape = true;
}


void gameOver()
{
  myScreen.gText(50, 50, "GAME OVER", redColour, whiteColour);
  delay(2000);
}


void tetrisGame(void)
{

  // get clock
  const unsigned long now = millis();

  // display new shape
  if ( newShape )
  {
    waitForRelease(INPUT_WAIT_NEW_SHAPE);
    newShape = false;

    // a new shape enters the game
    chooseNewShape();

    // clear next box
    myScreen.setPenSolid(true);
    myScreen.dRectangle(newBlockx + BLOCK_SIZE/2, newBlocky + BLOCK_SIZE,  BLOCK_SIZE*4,  BLOCK_SIZE*3, whiteColour);

    uint8_t *shape = allShapes[nextSelected[nextShapeNumber]];
    for ( int i = 0; i < 4; i++ )
    {
      uint8_t *block = shape + i * 2;
      myScreen.dRectangle(
        newBlockx + BLOCK_SIZE/2 + block[0]*BLOCK_SIZE,
        newBlocky + BLOCK_SIZE + block[1]*BLOCK_SIZE,
        BLOCK_SIZE - 2,
        BLOCK_SIZE - 2 ,
        colors[nextSelected[nextShapeNumber]]);
    }

    // check if new shape is placed over other shape(s)
    // on the board
    if ( checkEdge(0, 0, 0 ))
    {
      // draw shape to screen
      drawSelectedShape();
      gameOver();

      setup();
    }

    // draw shape to screen
    drawSelectedShape();
  }
  else
  {

    if ( now - lastDrop > dropWait )
    {
      // update clock
      lastDrop = now;
      moveShapeDown();
    }
  }

  if (!newShape)
  {
    userInput(now);
  }

}

void chooseNewShape()
{
  current = nextSelected[nextShapeNumber];

  nextShapeNumber++;

  if ( nextShapeNumber == 7 )
  {
    randomShape();
  }
  rot = 0;
  y = 0;
  x = BOARD_WIDTH / 2;

  old.rot = rot;
  old.y = y;
  old.x = x;
}

void userInput(unsigned long now)
{
  unsigned long waited = now - lastInput;
  unsigned long waitedY = now - lastInput;

  int jx =  getX();

  unsigned long move = INPUT_WAIT_MOVE / jx;

  if ( jx <  NEUTRAL && waited > -move )
  {

    if ( x < BOARD_WIDTH && !checkEdge(1, 0, 0))
    {
      x++;
      lastInput = now;
      drawSelectedShape();
      state = false;
    }
  }
  else if ( jx >  NEUTRAL && waited > move )
  {
    if  (x > 0 && !checkEdge(-1, 0, 0))
    {
      x--;
      lastInput = now;
      drawSelectedShape();
      state = false;
    }
  }

  int yRead = getY();   
  unsigned long moveY = INPUT_WAIT_MOVE / yRead;
  
  if ( yRead > SOFT && waitedY > moveY)
  {
    if  ( !checkEdge(0, 1, 0))
    {
      y++;
      lastInput = now;
      drawSelectedShape();
      state = false;
    }
  }

  if (waited > INPUT_WAIT_ROT)
  {
    if ( state && !checkEdge(0, 0, 1))
    {

      rot++;
      rot %= shapes[current];

      drawSelectedShape();
      lastInput = now;
      state = false;
    }
  }
}

void moveShapeDown()
{

  //Check Space Available
  if ( checkEdge(0, 1, 0 ))
  {
    //Shape already Below, Go to new Shape
    newShape = true;
    //Draw Shape
    uint8_t *shape = allShapes[current];
    for ( int i = 0; i < 4; i++ )
    {
      uint8_t *block = (shape + (rot * 4 + i) * 2);
      board[block[1] + y][block[0] + x] = current + 1;
    }
    lineCheck();
  }
  else
  {
    // Move Shape One Step Down
    y += 1;
    drawSelectedShape();
  }
}

void drawSelectedShape()
{
  uint8_t *shape = allShapes[current];
  for ( int i = 0; i < 4; i++ )
  {
    uint8_t *block = (shape + (rot * 4 + i) * 2);
    myScreen.setPenSolid(true);
    myScreen.dRectangle(
      BOARD_LEFT + block[0]*BLOCK_SIZE + BLOCK_SIZE * x,
      BOARD_TOP + block[1]*BLOCK_SIZE + BLOCK_SIZE * y,
      BLOCK_SIZE - 2,
      BLOCK_SIZE - 2 ,
      colors[current]);

    board[block[1] + y][block[0] + x] = 255;
  }

  // Remove Old Shape
  for ( int i = 0; i < 4; i++ )
  {
    uint8_t *block = (shape + (old.rot * 4 + i) * 2);

    if ( board[block[1] + old.y][block[0] + old.x] == 255 )
      continue;
    myScreen.setPenSolid(true);
    myScreen.dRectangle(
      BOARD_LEFT + block[0]*BLOCK_SIZE + BLOCK_SIZE * old.x,
      BOARD_TOP + block[1]*BLOCK_SIZE + BLOCK_SIZE * old.y,
      BLOCK_SIZE - 2,
      BLOCK_SIZE - 2 ,
      blackColour);
  }

  for ( int i = 0; i < 4; i++ )
  {
    uint8_t *block = (shape + (rot * 4 + i) * 2);
    board[block[1] + y][block[0] + x] = 0;
  }

  old.x = x;
  old.y = y;
  old.rot = rot;
}

boolean checkEdge(int xPosition, int yPosition, int rotation)
{
  uint8_t *shape = allShapes[current];
  for ( int i = 0; i < 4; i++ )
  {
    uint8_t *block = (shape + (((rot + rotation) % shapes[current]) * 4 + i) * 2);

    int newX = x + block[0] + xPosition;
    int newY = y + block[1] + yPosition;

    if ( newY == BOARD_HEIGHT || newX == BOARD_WIDTH || board[newY][newX] != 0 )
    {
      return true;
    }
  }
  return false;
}


void randomShape()
{
  //Randomize Select Shape from 7 Shapes

  for ( uint8_t i = 0; i < 7; i ++)
  {
    boolean sameShape;
    uint8_t newShape;
    do
    {
      newShape = random(7);

      // check if already in selected

      sameShape = false;
      for ( int j = 0; j < i; j++)
      {
        if ( newShape == nextSelected[j] )
        {
          sameShape = true;
          break;
        }
      }

    } while (sameShape);
    nextSelected[i] = newShape;
  }
  nextShapeNumber = 0;
}

void lineCheck()
{
  int currentLine;
  if ( y + 3 >= BOARD_HEIGHT )
  {
    currentLine = BOARD_HEIGHT - 1;
  }

  // scan board from current position
  for (int l = y; l <= currentLine; l++)
  {
    // check if there is a complete line on the board
    boolean line = true;

    for ( int c = 0; c < BOARD_WIDTH; c++ )
    {
      if (board[l][c] == 0)
      {
        line = false;
        break;
      }
    }

    if ( !line )
    {
      // move to next line
      continue;
    }

    lines++;
    playLineClear();
    
    if (lines == 3)
    {
      dropWait = DROP_WAIT_INIT_1;
      level = 1;
    }
    if (lines == 6)
    {
      dropWait = DROP_WAIT_INIT_2;
      level = 2;
    }

    if (lines == 9)
    {
      dropWait = DROP_WAIT_INIT_3;
      level = 3;
    }


    if (lines > highScore)
      highScore = lines;

    pointsCount();

    // move board down
    for ( int row = l; row > 0; row -- )
    {
      for ( int c = 0; c < BOARD_WIDTH; c++ )
      {
        uint8_t v = board[row - 1][c];

        board[row][c] = v;
        myScreen.setPenSolid(true);
        myScreen.dRectangle(
          BOARD_LEFT + BLOCK_SIZE * c,
          BOARD_TOP + BLOCK_SIZE * row,
          BLOCK_SIZE - 2,
          BLOCK_SIZE - 2 ,
          v == 0 ? blackColour : colors[v - 1] ) ;
      }
    }
  }

  delay(350);
}


void pointsCount()
{
  myScreen.gText(newBlockx + 2, 20, String(lines),  redColour, whiteColour);

  myScreen.gText(newBlockx + 2, 40, String(level),  redColour, whiteColour);
}


void pause()
{
  // check if the pushbutton is pressed.
  // if it is, the buttonState is HIGH:
  if (isPaused)
  {
    myScreen.gText(newBlockx, newBlocky + 40, "PAUSE", blackColour, whiteColour);
    //turn on blue led to indicate pause
    digitalWrite(BLUE_LED, HIGH);  
    while (isPaused)
    {
      delay(200);
      //wait until isPaused change State
    }
    myScreen.dRectangle(newBlockx, newBlocky + 40, 30, 10, blueColour);
    //turn off blue led to indicate pause
    digitalWrite(BLUE_LED, LOW);   
  }
}

